/*
  +----------------------------------------------------------------------+
  | teds extension for PHP                                               |
  | See COPYING file for further copyright information                   |
  +----------------------------------------------------------------------+
  | Author: Tyson Andre <tandre@php.net>                                 |
  +----------------------------------------------------------------------+
*/

/* This is a binary search tree. */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "zend_exceptions.h"

#include "php_teds.h"
#include "teds_sortedstrictset_arginfo.h"
#include "teds_sortedstrictset.h"
#include "teds_util.h"
#include "teds.h"
// #include "ext/spl/spl_functions.h"
#include "ext/spl/spl_engine.h"
#include "ext/spl/spl_exceptions.h"
#include "ext/spl/spl_iterators.h"
#include "ext/json/php_json.h"

#include <stdbool.h>

zend_object_handlers teds_handler_SortedStrictSet;
zend_class_entry *teds_ce_SortedStrictSet;

static zend_always_inline teds_sortedstrictset_node *teds_sortedstrictset_tree_find_key(const teds_sortedstrictset_tree *tree, zval *key)
{
	teds_sortedstrictset_node *it = tree->root;
	while (it != NULL) {
		const int comparison = teds_stable_compare(key, &it->key);
		if (comparison > 0) {
			it = it->right;
		} else if (comparison < 0) {
			it = it->left;
		} else {
			return it;
		}
	}
	return NULL;
}

static zend_always_inline void teds_sortedstrictset_tree_remove_node(teds_sortedstrictset_tree *array, teds_sortedstrictset_node *const node, bool free_zval);

static zend_always_inline teds_sortedstrictset_node *teds_sortedstrictset_node_alloc(const zval *key, teds_sortedstrictset_node *parent) {
	teds_sortedstrictset_node *n = emalloc(sizeof(teds_sortedstrictset_node));
	n->parent = parent;
	ZVAL_COPY(&n->key, key);
	TEDS_SORTEDSTRICTSET_NODE_REFCOUNT(n) = 1;
	return n;
}

static zend_always_inline void teds_sortedstrictset_node_release(teds_sortedstrictset_node *node) {
	ZEND_ASSERT(node != NULL);
	ZEND_ASSERT(TEDS_SORTEDSTRICTSET_NODE_REFCOUNT(node) > 0);
	if (--TEDS_SORTEDSTRICTSET_NODE_REFCOUNT(node) == 0) {
		ZEND_ASSERT(Z_ISUNDEF(node->key));
		efree_size(node, sizeof(teds_sortedstrictset_node));
	}
}

/* Returns true if a new entry was added to the map, false if updated. Based on _zend_hash_add_or_update_i. */
static zend_always_inline bool teds_sortedstrictset_tree_insert(teds_sortedstrictset_tree *tree, zval *key, bool add_new)
{
	ZEND_ASSERT(Z_TYPE_P(key) != IS_UNDEF);

	/* TODO optimize */
	teds_sortedstrictset_node *it = tree->root;
	if (it == NULL) {
		/* Initialize this tree as a new binary search tree of size 1 */
		it = teds_sortedstrictset_node_alloc(key, NULL);
		tree->root = it;
		it->left = NULL;
		it->right = NULL;
		it->prev = NULL;
		it->next = NULL;
		tree->nNumOfElements++;
		return true;
	}
	while (true) {
		const int comparison = teds_stable_compare(key, &it->key);
		if (comparison > 0) {
			if (it->right == NULL) {
				teds_sortedstrictset_node *const c = teds_sortedstrictset_node_alloc(key, it);
				teds_sortedstrictset_node *const next = it->next;

				c->left = NULL;
				c->right = NULL;
				c->prev = it;
				c->next = next;
				it->next = c;
				it->right = c;
				if (next) {
					next->prev = c;
				}
				tree->nNumOfElements++;
				/* TODO rebalance */

				return true;
			}
			it = it->right;
		} else if ((add_new && !ZEND_DEBUG) || comparison < 0) {
			if (it->left == NULL) {
				teds_sortedstrictset_node *const c = teds_sortedstrictset_node_alloc(key, it);
				teds_sortedstrictset_node *const prev = it->prev;

				c->left = NULL;
				c->right = NULL;
				c->prev = prev;
				c->next = it;
				it->prev = c;
				it->left = c;
				if (prev) {
					prev->next = c;
				}
				tree->nNumOfElements++;
				/* TODO rebalance */

				return true;

			}
			it = it->left;
		} else {
			ZEND_ASSERT(!add_new);
			/* Already exists */
			return false;
		}
	}
}

static void teds_sortedstrictset_tree_clear(teds_sortedstrictset_tree *array);

/* Used by InternalIterator returned by SortedStrictSet->getIterator() */
typedef struct _teds_sortedstrictset_it {
	zend_object_iterator       intern;
	teds_sortedstrictset_node *node;
} teds_sortedstrictset_it;

static zend_always_inline teds_sortedstrictset *teds_sortedstrictset_from_obj(zend_object *obj)
{
	return (teds_sortedstrictset*)((char*)(obj) - XtOffsetOf(teds_sortedstrictset, std));
}

#define Z_SORTEDSTRICTSET_P(zv)  teds_sortedstrictset_from_obj(Z_OBJ_P((zv)))

static bool teds_sortedstrictset_tree_uninitialized(teds_sortedstrictset_tree *array)
{
	if (array->initialized) {
		return false;
	}
	ZEND_ASSERT(array->root == NULL);
	ZEND_ASSERT(array->nNumOfElements == 0);
	return true;
}

static void teds_sortedstrictset_tree_set_empty_tree(teds_sortedstrictset_tree *array)
{
	array->root = NULL;
	array->nNumOfElements = 0;
	array->initialized = true;
}

void teds_sortedstrictset_tree_init_from_array(teds_sortedstrictset_tree *array, zend_array *values)
{
	zval *val;

	teds_sortedstrictset_tree_set_empty_tree(array);
	ZEND_HASH_FOREACH_VAL(values, val)  {
		ZVAL_DEREF(val);
		teds_sortedstrictset_tree_insert(array, val, false);
	} ZEND_HASH_FOREACH_END();
}

void teds_sortedstrictset_tree_init_from_traversable(teds_sortedstrictset_tree *array, zend_object *obj)
{
	zend_class_entry *ce = obj->ce;
	zend_object_iterator *iter;
	teds_sortedstrictset_tree_set_empty_tree(array);
	zval tmp_obj;
	ZVAL_OBJ(&tmp_obj, obj);
	iter = ce->get_iterator(ce, &tmp_obj, 0);

	if (UNEXPECTED(EG(exception))) {
		return;
	}

	const zend_object_iterator_funcs *funcs = iter->funcs;

	if (funcs->rewind) {
		funcs->rewind(iter);
		if (UNEXPECTED(EG(exception))) {
			return;
		}
	}

	while (funcs->valid(iter) == SUCCESS) {
		if (UNEXPECTED(EG(exception))) {
			break;
		}
		zval *value = funcs->get_current_data(iter);
		if (UNEXPECTED(EG(exception)) || value == NULL) {
			break;
		}

		ZVAL_DEREF(value);
		const bool created_new_entry = teds_sortedstrictset_tree_insert(array, value, false);
		if (!created_new_entry) {
			if (UNEXPECTED(EG(exception))) {
				break;
			}
		}

		iter->index++;
		funcs->move_forward(iter);
		if (UNEXPECTED(EG(exception))) {
			break;
		}
	}

	if (iter) {
		zend_iterator_dtor(iter);
	}
	if (UNEXPECTED(EG(exception))) {
		teds_sortedstrictset_tree_clear(array);
	}
}

static teds_sortedstrictset_node *teds_sortedstrictset_node_copy_ctor_recursive(const teds_sortedstrictset_node *from, teds_sortedstrictset_node *parent, teds_sortedstrictset_node *left_parent_node, teds_sortedstrictset_node *right_parent_node) {
	ZEND_ASSERT(from != NULL);
	teds_sortedstrictset_node *copy = teds_sortedstrictset_node_alloc(&from->key, parent);
	if (from->left) {
		copy->left = teds_sortedstrictset_node_copy_ctor_recursive(from->left, copy, left_parent_node, copy);
	} else {
		copy->left = NULL;
		/* This is the first node after left_parent_node */
		copy->prev = left_parent_node;
		if (left_parent_node) {
			left_parent_node->next = copy;
		}
	}
	if (from->right) {
		copy->right = teds_sortedstrictset_node_copy_ctor_recursive(from->right, copy, copy, right_parent_node);
	} else {
		copy->right = NULL;
		/* This is the last node before right_parent_node */
		copy->next = right_parent_node;
		if (right_parent_node) {
			right_parent_node->prev = copy;
		}
	}
	return copy;
}

static void teds_sortedstrictset_tree_copy_ctor(teds_sortedstrictset_tree *to, teds_sortedstrictset_tree *from)
{
	teds_sortedstrictset_tree_set_empty_tree(to);
	/* Copy the original tree structure. It will be balanced if the original tree is balanced. */
	to->nNumOfElements = from->nNumOfElements;
	to->initialized = true;
	if (!teds_sortedstrictset_tree_empty_size(from)) {
		to->root = teds_sortedstrictset_node_copy_ctor_recursive(from->root, NULL, NULL, NULL);
	} else {
		to->root = NULL;
	}
}

static void teds_sortedstrictset_node_dtor(teds_sortedstrictset_node *node)
{
	/* Free keys in sorted order */
	while (node != NULL) {
		teds_sortedstrictset_node_dtor(node->left);
		teds_sortedstrictset_node *right = node->right;
		zval_ptr_dtor(&node->key);
		ZVAL_UNDEF(&node->key);
		teds_sortedstrictset_node_release(node);
		node = right;
	}
}

/* Destructs and frees contents and the array itself.
 * If you want to re-use the array then you need to re-initialize it.
 */
void teds_sortedstrictset_tree_dtor(teds_sortedstrictset_tree *array)
{
	if (teds_sortedstrictset_tree_empty_size(array)) {
		return;
	}
	teds_sortedstrictset_node *root = array->root;
	teds_sortedstrictset_tree_set_empty_tree(array);
	teds_sortedstrictset_node_dtor(root);
}

static HashTable* teds_sortedstrictset_get_gc(zend_object *obj, zval **table, int *table_count)
{
	teds_sortedstrictset *intern = teds_sortedstrictset_from_obj(obj);
	if (intern->array.nNumOfElements > 0) {
		zend_get_gc_buffer *gc_buffer = zend_get_gc_buffer_create();
		zval *key;

		TEDS_SORTEDSTRICTSET_FOREACH_KEY(&intern->array, key) {
			zend_get_gc_buffer_add_zval(gc_buffer, key);
		} TEDS_SORTEDSTRICTSET_FOREACH_END();

		zend_get_gc_buffer_use(gc_buffer, table, table_count);
	}

	// Returning the object's properties is redundant if dynamic properties are not allowed,
	// and this can't be subclassed.
	return NULL;
}

static HashTable* teds_sortedstrictset_get_properties(zend_object *obj)
{
	teds_sortedstrictset *intern = teds_sortedstrictset_from_obj(obj);
	const uint32_t len = intern->array.nNumOfElements;
	HashTable *ht = zend_std_get_properties(obj);
	uint32_t old_length = zend_hash_num_elements(ht);
	/* Initialize properties array */
	uint32_t i = 0;
	zval *key;
	TEDS_SORTEDSTRICTSET_FOREACH_KEY(&intern->array, key) {
		Z_TRY_ADDREF_P(key);
		zend_hash_index_update(ht, i, key);
		i++;
	} TEDS_SORTEDSTRICTSET_FOREACH_END();

	ZEND_ASSERT(i == len);

	for (size_t i = len; i < old_length; i++) {
		zend_hash_index_del(ht, i);
	}

	return ht;
}

static void teds_sortedstrictset_free_storage(zend_object *object)
{
	teds_sortedstrictset *intern = teds_sortedstrictset_from_obj(object);
	teds_sortedstrictset_tree_dtor(&intern->array);
	zend_object_std_dtor(&intern->std);
}

static zend_object *teds_sortedstrictset_new_ex(zend_class_entry *class_type, zend_object *orig, bool clone_orig)
{
	teds_sortedstrictset *intern;

	intern = zend_object_alloc(sizeof(teds_sortedstrictset), class_type);
	/* This is a final class */
	ZEND_ASSERT(class_type == teds_ce_SortedStrictSet);

	zend_object_std_init(&intern->std, class_type);
	object_properties_init(&intern->std, class_type);
	intern->std.handlers = &teds_handler_SortedStrictSet;

	if (orig && clone_orig) {
		teds_sortedstrictset *other = teds_sortedstrictset_from_obj(orig);
		teds_sortedstrictset_tree_copy_ctor(&intern->array, &other->array);
	} else {
		intern->array.root = NULL;
	}

	return &intern->std;
}

static zend_object *teds_sortedstrictset_new(zend_class_entry *class_type)
{
	return teds_sortedstrictset_new_ex(class_type, NULL, 0);
}


static zend_object *teds_sortedstrictset_clone(zend_object *old_object)
{
	zend_object *new_object = teds_sortedstrictset_new_ex(old_object->ce, old_object, 1);

	zend_objects_clone_members(new_object, old_object);

	return new_object;
}

static int teds_sortedstrictset_count_elements(zend_object *object, zend_long *count)
{
	teds_sortedstrictset *intern;

	intern = teds_sortedstrictset_from_obj(object);
	*count = intern->array.nNumOfElements;
	return SUCCESS;
}

/* Get number of entries in this SortedStrictSet */
PHP_METHOD(Teds_SortedStrictSet, count)
{
	zval *object = ZEND_THIS;

	ZEND_PARSE_PARAMETERS_NONE();

	teds_sortedstrictset *intern = Z_SORTEDSTRICTSET_P(object);
	RETURN_LONG(intern->array.nNumOfElements);
}

/* Get whether this SortedStrictSet is empty */
PHP_METHOD(Teds_SortedStrictSet, isEmpty)
{
	zval *object = ZEND_THIS;

	ZEND_PARSE_PARAMETERS_NONE();

	teds_sortedstrictset *intern = Z_SORTEDSTRICTSET_P(object);
	RETURN_BOOL(intern->array.nNumOfElements == 0);
}

/* Create this from an iterable */
PHP_METHOD(Teds_SortedStrictSet, __construct)
{
	zval* iterable = NULL;

	ZEND_PARSE_PARAMETERS_START(0, 1)
		Z_PARAM_OPTIONAL
		Z_PARAM_ITERABLE(iterable)
	ZEND_PARSE_PARAMETERS_END();

	teds_sortedstrictset *intern = Z_SORTEDSTRICTSET_P(ZEND_THIS);

	if (UNEXPECTED(!teds_sortedstrictset_tree_uninitialized(&intern->array))) {
		zend_throw_exception(spl_ce_RuntimeException, "Called Teds\\SortedStrictSet::__construct twice", 0);
		/* called __construct() twice, bail out */
		RETURN_THROWS();
	}

	if (iterable == NULL) {
		teds_sortedstrictset_tree_set_empty_tree(&intern->array);
		return;
	}

	switch (Z_TYPE_P(iterable)) {
		case IS_ARRAY:
			teds_sortedstrictset_tree_init_from_array(&intern->array, Z_ARRVAL_P(iterable));
			return;
		case IS_OBJECT:
			teds_sortedstrictset_tree_init_from_traversable(&intern->array, Z_OBJ_P(iterable));
			return;
		EMPTY_SWITCH_DEFAULT_CASE();
	}
}

PHP_METHOD(Teds_SortedStrictSet, getIterator)
{
	ZEND_PARSE_PARAMETERS_NONE();

	zend_create_internal_iterator_zval(return_value, ZEND_THIS);
}

static void teds_sortedstrictset_it_dtor(zend_object_iterator *iter)
{
	teds_sortedstrictset_node *node = ((teds_sortedstrictset_it*)iter)->node;
	if (node) {
		teds_sortedstrictset_node_release(node);
		((teds_sortedstrictset_it*)iter)->node = NULL;
	}
	zval_ptr_dtor(&iter->data);
}

static void teds_sortedstrictset_it_rewind(zend_object_iterator *iter)
{
	teds_sortedstrictset *object   = Z_SORTEDSTRICTSET_P(&iter->data);
	teds_sortedstrictset_node *const orig_node = ((teds_sortedstrictset_it*)iter)->node;
	teds_sortedstrictset_node *const new_node = teds_sortedstrictset_tree_get_first(&object->array);
	if (new_node == orig_node) {
		return;
	}
	((teds_sortedstrictset_it*)iter)->node = new_node;
	if (new_node != NULL) {
		TEDS_SORTEDSTRICTSET_NODE_REFCOUNT(new_node)++;
	}
	if (orig_node != NULL) {
		teds_sortedstrictset_node_release(orig_node);
	}
}

static zend_always_inline bool teds_sortedstrictset_node_valid(teds_sortedstrictset_node *node) {
	/* TODO: Mark key as invalid when removing? */
	return node != NULL && !Z_ISUNDEF(node->key);
}

static int teds_sortedstrictset_it_valid(zend_object_iterator *iter)
{
	teds_sortedstrictset_it     *iterator = (teds_sortedstrictset_it*)iter;
	return teds_sortedstrictset_node_valid(iterator->node) ? SUCCESS : FAILURE;
}

static zval *teds_sortedstrictset_it_get_current_data(zend_object_iterator *iter)
{
	teds_sortedstrictset_node *node = ((teds_sortedstrictset_it*)iter)->node;
	if (teds_sortedstrictset_node_valid(node)) {
		return &node->key;
	} else {
		return &EG(uninitialized_zval);
	}
}

static void teds_sortedstrictset_it_get_current_key(zend_object_iterator *iter, zval *key)
{
	teds_sortedstrictset_node *node = ((teds_sortedstrictset_it*)iter)->node;
	if (teds_sortedstrictset_node_valid(node)) {
		ZVAL_COPY(key, &node->key);
	} else {
		ZVAL_NULL(key);
	}
}

static void teds_sortedstrictset_it_move_forward(zend_object_iterator *iter)
{
	teds_sortedstrictset_node *const node = ((teds_sortedstrictset_it*)iter)->node;
	if (!teds_sortedstrictset_node_valid(node)) {
		return;
	}
	teds_sortedstrictset_node *const next = node->next;
	((teds_sortedstrictset_it*)iter)->node = next;
	if (next) {
		TEDS_SORTEDSTRICTSET_NODE_REFCOUNT(next)++;
	}
	teds_sortedstrictset_node_release(node);
}

/* iterator handler table */
static const zend_object_iterator_funcs teds_sortedstrictset_it_funcs = {
	teds_sortedstrictset_it_dtor,
	teds_sortedstrictset_it_valid,
	teds_sortedstrictset_it_get_current_data,
	teds_sortedstrictset_it_get_current_key,
	teds_sortedstrictset_it_move_forward,
	teds_sortedstrictset_it_rewind,
	NULL,
	NULL, /* get_gc */
};


zend_object_iterator *teds_sortedstrictset_get_iterator(zend_class_entry *ce, zval *object, int by_ref)
{
	teds_sortedstrictset_it *iterator;

	if (UNEXPECTED(by_ref)) {
		zend_throw_error(NULL, "An iterator cannot be used with foreach by reference");
		return NULL;
	}

	iterator = emalloc(sizeof(teds_sortedstrictset_it));

	zend_iterator_init((zend_object_iterator*)iterator);

	ZVAL_OBJ_COPY(&iterator->intern.data, Z_OBJ_P(object));
	iterator->intern.funcs = &teds_sortedstrictset_it_funcs;
	teds_sortedstrictset_node *node = teds_sortedstrictset_tree_get_first(&Z_SORTEDSTRICTSET_P(object)->array);
	if (node) {
		TEDS_SORTEDSTRICTSET_NODE_REFCOUNT(node)++;
	}
	iterator->node = node;
	(void) ce;

	return &iterator->intern;
}

PHP_METHOD(Teds_SortedStrictSet, __unserialize)
{
	HashTable *raw_data;

	if (zend_parse_parameters(ZEND_NUM_ARGS(), "h", &raw_data) == FAILURE) {
		RETURN_THROWS();
	}

	teds_sortedstrictset *const intern = Z_SORTEDSTRICTSET_P(ZEND_THIS);
	teds_sortedstrictset_tree *const array = &intern->array;
	if (UNEXPECTED(!teds_sortedstrictset_tree_uninitialized(array))) {
		zend_throw_exception(spl_ce_RuntimeException, "Already unserialized", 0);
		RETURN_THROWS();
	}

	teds_sortedstrictset_tree_set_empty_tree(array);

	zend_string *str;
	zval *val;

	ZEND_HASH_FOREACH_STR_KEY_VAL(raw_data, str, val) {
		if (UNEXPECTED(str)) {
			teds_sortedstrictset_tree_clear(array);
			zend_throw_exception(spl_ce_UnexpectedValueException, "Teds\\SortedStrictSet::__unserialize saw unexpected string key, expected sequence of values", 0);
			RETURN_THROWS();
		}

		teds_sortedstrictset_tree_insert(array, val, false);
	} ZEND_HASH_FOREACH_END();
}

PHP_METHOD(Teds_SortedStrictSet, __set_state)
{
	zend_array *array_ht;

	ZEND_PARSE_PARAMETERS_START(1, 1)
		Z_PARAM_ARRAY_HT(array_ht)
	ZEND_PARSE_PARAMETERS_END();
	zend_object *object = teds_sortedstrictset_new(teds_ce_SortedStrictSet);
	teds_sortedstrictset *intern = teds_sortedstrictset_from_obj(object);
	teds_sortedstrictset_tree_init_from_array(&intern->array, array_ht);

	RETURN_OBJ(object);
}

#define IMPLEMENT_READ_POSITION_PHP_METHOD(methodName, pos) \
PHP_METHOD(Teds_SortedStrictSet, methodName) \
{ \
       ZEND_PARSE_PARAMETERS_NONE(); \
       const teds_sortedstrictset *intern = Z_SORTEDSTRICTSET_P(ZEND_THIS); \
       if (teds_sortedstrictset_tree_empty_size(&intern->array)) { \
               zend_throw_exception(spl_ce_UnderflowException, "Cannot read " # methodName " of empty SortedStrictSet", 0); \
               RETURN_THROWS(); \
       } \
       teds_sortedstrictset_node *node = teds_sortedstrictset_tree_get_ ## pos(&intern->array); \
       RETVAL_COPY(&node->key); \
}

IMPLEMENT_READ_POSITION_PHP_METHOD(bottom, first)
IMPLEMENT_READ_POSITION_PHP_METHOD(top, last)

#define IMPLEMENT_REMOVE_POSITION_PHP_METHOD(methodName, pos) \
PHP_METHOD(Teds_SortedStrictSet, methodName) \
{ \
	ZEND_PARSE_PARAMETERS_NONE(); \
	teds_sortedstrictset *intern = Z_SORTEDSTRICTSET_P(ZEND_THIS); \
	if (teds_sortedstrictset_tree_empty_size(&intern->array)) { \
		zend_throw_exception(spl_ce_UnderflowException, "Cannot " # methodName " from empty SortedStrictSet", 0); \
		RETURN_THROWS(); \
	} \
	teds_sortedstrictset_node *node = teds_sortedstrictset_tree_get_ ## pos(&intern->array); \
	RETVAL_COPY_VALUE(&node->key); \
	teds_sortedstrictset_tree_remove_node(&intern->array, node, false); \
}

IMPLEMENT_REMOVE_POSITION_PHP_METHOD(pop, last)
IMPLEMENT_REMOVE_POSITION_PHP_METHOD(shift, first)

PHP_METHOD(Teds_SortedStrictSet, values)
{
	ZEND_PARSE_PARAMETERS_NONE();
	teds_sortedstrictset *intern = Z_SORTEDSTRICTSET_P(ZEND_THIS);
	size_t len = intern->array.nNumOfElements;
	if (!len) {
		RETURN_EMPTY_ARRAY();
	}
	zend_array *values = zend_new_array(len);
	/* Initialize return array */
	zend_hash_real_init_packed(values);

	/* Go through values and add values to the return array */
	ZEND_HASH_FILL_PACKED(values) {
		zval *value;
		TEDS_SORTEDSTRICTSET_FOREACH_KEY(&intern->array, value) {
			Z_TRY_ADDREF_P(value);
			ZEND_HASH_FILL_ADD(value);
		} TEDS_SORTEDSTRICTSET_FOREACH_END();
	} ZEND_HASH_FILL_END();
	RETURN_ARR(values);
}

static void teds_sortedstrictset_tree_replace_node_with_child(teds_sortedstrictset_tree *tree, teds_sortedstrictset_node *removed_node, teds_sortedstrictset_node *child_node) {
	teds_sortedstrictset_node *const parent = removed_node->parent;
	if (child_node) {
		child_node->parent = parent;
	}
	if (parent == NULL) {
		ZEND_ASSERT(removed_node == tree->root);
		tree->root = child_node;
		return;
	}
	if (parent->left == removed_node) {
		parent->left = child_node;
	} else {
		ZEND_ASSERT(parent->right == removed_node);
		parent->right = child_node;
	}
}

static zend_always_inline teds_sortedstrictset_node *teds_sortedstrictset_node_remove_leftmost(teds_sortedstrictset_node *node) {
	while (true) {
		ZEND_ASSERT(node != NULL);
		ZEND_ASSERT(node->parent != NULL);
		if (node->left) {
			node = node->left;
			continue;
		}
		teds_sortedstrictset_node *right = node->right;
		if (right) {
			right->parent = node->parent;
			if (node->parent->left == node) {
				node->parent->left = right;
			} else {
				ZEND_ASSERT(node->parent->right == node);
				node->parent->right = right;
			}
			node->right = NULL;
		}
		node->parent = NULL;
		return node;
	}
}

static zend_always_inline void teds_sortedstrictset_tree_remove_node(teds_sortedstrictset_tree *tree, teds_sortedstrictset_node *const node, bool free_zval) {
	teds_sortedstrictset_node *const prev = node->prev;
	teds_sortedstrictset_node *const next = node->next;
	if (prev) {
		prev->next = next;
	}
	if (next) {
		next->prev = prev;
	}
	node->prev = NULL;
	node->next = NULL;

	if (!node->left) {
		teds_sortedstrictset_tree_replace_node_with_child(tree, node, node->right);
	} else if (!node->right) {
		teds_sortedstrictset_tree_replace_node_with_child(tree, node, node->left);
	} else {
		teds_sortedstrictset_node *const replacement = teds_sortedstrictset_node_remove_leftmost(node->right);
		teds_sortedstrictset_node *const parent = node->parent;
		replacement->left = node->left;
		if (node->left) {
			node->left->parent = replacement;
		}
		replacement->right = node->right;
		if (node->right) {
			node->right->parent = replacement;
		}
		replacement->parent = parent;
		if (parent == NULL) {
			ZEND_ASSERT(tree->root == node);
			tree->root = replacement;
		} else {
			if (parent->left == node) {
				parent->left = replacement;
			} else {
				ZEND_ASSERT(parent->right == node);
				parent->right = replacement;
			}
		}
	}

	ZEND_ASSERT(tree->nNumOfElements > 0);
	tree->nNumOfElements--;
	if (free_zval) {
		zval_ptr_dtor(&node->key);
	}
	ZVAL_UNDEF(&node->key);
	teds_sortedstrictset_node_release(node);
}

static bool teds_sortedstrictset_tree_remove_key(teds_sortedstrictset_tree *tree, zval *key)
{
	/* FIXME implement binary tree removal */
	teds_sortedstrictset_node *const node = teds_sortedstrictset_tree_find_key(tree, key);
	if (node == NULL) {
		/* Nothing to remove */
		return false;
	}
	teds_sortedstrictset_tree_remove_node(tree, node, true);
	return true;
}

PHP_METHOD(Teds_SortedStrictSet, add)
{
	zval *value;
	ZEND_PARSE_PARAMETERS_START(1, 1)
		Z_PARAM_ZVAL(value)
	ZEND_PARSE_PARAMETERS_END();

	teds_sortedstrictset *intern = Z_SORTEDSTRICTSET_P(ZEND_THIS);
	RETURN_BOOL(teds_sortedstrictset_tree_insert(&intern->array, value, false));
}

PHP_METHOD(Teds_SortedStrictSet, remove)
{
	zval *key;
	ZEND_PARSE_PARAMETERS_START(1, 1)
		Z_PARAM_ZVAL(key)
	ZEND_PARSE_PARAMETERS_END();

	teds_sortedstrictset *intern = Z_SORTEDSTRICTSET_P(ZEND_THIS);
	RETURN_BOOL(teds_sortedstrictset_tree_remove_key(&intern->array, key));
}

PHP_METHOD(Teds_SortedStrictSet, contains)
{
	zval *key;
	ZEND_PARSE_PARAMETERS_START(1, 1)
		Z_PARAM_ZVAL(key)
	ZEND_PARSE_PARAMETERS_END();

	const teds_sortedstrictset *intern = Z_SORTEDSTRICTSET_P(ZEND_THIS);
	if (intern->array.nNumOfElements > 0) {
		teds_sortedstrictset_node *entry = teds_sortedstrictset_tree_find_key(&intern->array, key);
		RETURN_BOOL(entry != NULL);
	}
	RETURN_FALSE;
}

static void teds_sortedstrictset_tree_clear(teds_sortedstrictset_tree *array) {
	if (teds_sortedstrictset_tree_empty_size(array)) {
		return;
	}
	teds_sortedstrictset_tree array_copy = *array;

	teds_sortedstrictset_tree_set_empty_tree(array);

	teds_sortedstrictset_tree_dtor(&array_copy);
	/* Could call teds_sortedstrictset_get_properties but properties array is typically not initialized unless var_dump or other inefficient functionality is used */
}

PHP_METHOD(Teds_SortedStrictSet, clear)
{
	ZEND_PARSE_PARAMETERS_NONE();
	teds_sortedstrictset *intern = Z_SORTEDSTRICTSET_P(ZEND_THIS);
	teds_sortedstrictset_tree_clear(&intern->array);
	TEDS_RETURN_VOID();
}

PHP_MINIT_FUNCTION(teds_sortedstrictset)
{
	TEDS_MINIT_IGNORE_UNUSED();
	teds_ce_SortedStrictSet = register_class_Teds_SortedStrictSet(zend_ce_aggregate, zend_ce_countable, php_json_serializable_ce);
	teds_ce_SortedStrictSet->create_object = teds_sortedstrictset_new;

	memcpy(&teds_handler_SortedStrictSet, &std_object_handlers, sizeof(zend_object_handlers));

	teds_handler_SortedStrictSet.offset          = XtOffsetOf(teds_sortedstrictset, std);
	teds_handler_SortedStrictSet.clone_obj       = teds_sortedstrictset_clone;
	teds_handler_SortedStrictSet.count_elements  = teds_sortedstrictset_count_elements;
	teds_handler_SortedStrictSet.get_properties  = teds_sortedstrictset_get_properties;
	teds_handler_SortedStrictSet.get_gc          = teds_sortedstrictset_get_gc;
	teds_handler_SortedStrictSet.dtor_obj        = zend_objects_destroy_object;
	teds_handler_SortedStrictSet.free_obj        = teds_sortedstrictset_free_storage;

	teds_ce_SortedStrictSet->ce_flags |= ZEND_ACC_FINAL | ZEND_ACC_NO_DYNAMIC_PROPERTIES;
	teds_ce_SortedStrictSet->get_iterator = teds_sortedstrictset_get_iterator;

	return SUCCESS;
}
